//
// "$Id: Fl_Level.cxx 5472 2006-09-20 03:03:14Z mike $"
//
// Level widget for the Fast Light Tool Kit (FLTK).
//
// Copyright 2008 by Kurt Van Dijck
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//

#include <FL/Fl.H>
#include <FL/fl_draw.H>
#include <stdlib.h>
#include <FL/math.h>

#include "Fl_Level.H"
#include <error.h>

// All angles are measured with 0 to the right and counter-clockwise

void Fl_Level::draw(int X, int Y, int W, int H) {
   fl_push_clip(X, Y, W, H);
   //if (damage()&FL_DAMAGE_ALL)
      draw_box(box(), X, Y, W, H, FL_BACKGROUND2_COLOR);
   X += Fl::box_dx(box());
   Y += Fl::box_dy(box());
   W -= Fl::box_dw(box());
   H -= Fl::box_dh(box());
   //double angle = (a2-a1)*(value()-minimum())/(maximum()-minimum()) + a1;

   if (type() == FL_1D) {
      if (W >= H) {
         int tp;
         int belw = (W/7) & ~1;
         int cw = belw +4;
         int belh = H *3/4;
         fl_draw_box(FL_FLAT_BOX, X +(W -cw)/2, Y, cw, H, color());
         if (active()) {
            if (value() < minimum()) {
               tp = 0;
            } else if (value() > maximum()) {
               tp = W;
            } else {
               tp = lround((value() - minimum()) *W/(maximum() - minimum()));
            }
            fl_draw_box(FL_ROUNDED_BOX, X +tp -belw/2, Y +(H-belh)/2, belw, belh, selection_color());
         }
      } else {
         int tp;
         int belh = (H/7) &~1;
         int ch = belh +4;
         int belw = W *3/4;
         fl_draw_box(FL_FLAT_BOX, X, Y +(H-ch)/2, W, ch, color());
         if (active()) {
            if (value() < minimum()) {
               tp = 0;
            } else if (value() > maximum()) {
               tp = H;
            } else {
               tp = lround((value() - minimum()) *H/(maximum() - minimum()));
            }
            fl_draw_box(FL_ROUNDED_BOX, X +(W -belw)/2, Y +tp -belh/2, belw, belh, selection_color());
         }
      }
   } else if (FL_2D == type()) {
      int bel = (W/10) & ~1;
      int cs = bel +6;
      int xc = X + W/2;
      int yc = Y + H/2;
      double x, y;
      fl_color(color());
      fl_pie(xc -cs , yc -cs , 2*cs , 2*cs , 0, 360);
      if (active()) {
         x = value();
         if (x < minimum())
            x = minimum();
         else if (x > maximum())
            x = maximum();
         y = yvalue();
         if (y < minimum())
            y = minimum();
         else if (y > maximum())
            y = maximum();
       
         xc = X + lround((x - minimum())*W/(maximum() - minimum()));
         yc = Y + lround((y - minimum())*H/(maximum() - minimum()));
         fl_color(selection_color());
         fl_pie(xc -bel, yc -bel, 2*bel, 2*bel, 0, 360);
      }
   }
   fl_pop_clip();
}

void Fl_Level::draw() {
   draw(x(), y(), w(), h());
   draw_label();
}

int Fl_Level::handle(int event, int X, int Y, int W, int H) {
   switch (event) {
   case FL_PUSH:
      handle_push();
   case FL_DRAG:
      if (FL_1D == type()) {
         double val;
         if (W >= H) {
            val = (Fl::event_x() - X)*(maximum() - minimum())/W;
         } else {
            val = (Fl::event_y() - Y)*(maximum() - minimum())/H;
         }
         handle_drag(clamp(round(val)));
      }
      return 1;
   case FL_RELEASE:
      handle_release();
      return 1;
   case FL_ENTER :
   case FL_LEAVE :
      return 1;
   default:
      return 0;
   }
}

int Fl_Level::handle(int e) {
   return handle(e, x(), y(), w(), h());
}

Fl_Level::Fl_Level(int X, int Y, int W, int H, const char* l)
   :Fl_Valuator(X, Y, W, H, l) {
   box(FL_FLAT_BOX);
   color(FL_DARK_GREEN);
   selection_color(FL_MAGENTA);
   base::set_output();
   type(FL_1D);
   range(-1, +1);
   value(0, 0);
}

//
// End of "$Id: Fl_Level.cxx 5472 2006-09-20 03:03:14Z mike $".
//
