/* -*-  tab-width:4; c-basic-offset:4  -*- 
 * rotate.c -- determine Freerunner orientation.
 * Author   -- Chris Ball <cjb@laptop.org>
 *
 * This file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 */

#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <linux/input.h>
#include <error.h>

//#include <exception>
#include <Fl/Fl.H>
#include <FL/Fl_Group.H>
#include <FL/Fl_Valuator.H>
#include <FL/Fl_Menu_.H>
#include <FL/Fl_Button.H>

#include <FL/Fl_Light_Button.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Tabs.H>
#include <FL/Fl_Choice.H>
#include <FL/Fl_Menu.H>
#include <FL/Fl_Dial.H>
#include <FL/Fl_Value_Slider.H>
#include <FL/Fl_Slider.H>
#include <FL/Fl_Value_Output.H>
#include <FL/fl_ask.H>

#include "Fl_Level.H"

static const char dev0 [] = "/dev/input/event2";
static const char dev1 [] = "/dev/input/event3";

struct _3d {
   double x, y, z;
};
struct sens {
   int use;
   class Fl_Button *bttn;
   time_t time;
   struct _3d raw, cal;
};
struct statics {
   int sel;
   #define SEL_X  0
   #define SEL_Y  1
   #define SEL_Z  2
   #define SEL_XY 3
   #define SEL_YZ 4
   #define SEL_XZ 5
   class Fl_Level * wid;

   struct _3d a;
   struct sens s[2];
   class Fl_Widget *cal_bttn;
} s;

static void save_cal (void);
static void load_cal (void);

static void cb_calibrate (Fl_Widget *widget, void * v) {
   switch (s.sel) {
   case SEL_X:
      if (s.s[0].use)
         s.s[0].cal.x = -s.s[0].raw.x;
      if (s.s[1].use)
         s.s[1].cal.x = -s.s[1].raw.x;
      break;
   case SEL_Y:
      if (s.s[0].use)
         s.s[0].cal.y = -s.s[0].raw.y;
      if (s.s[1].use)
         s.s[1].cal.y = -s.s[1].raw.y;
      break;
   case SEL_Z:
      if (s.s[0].use)
         s.s[0].cal.z = -s.s[0].raw.z;
      if (s.s[1].use)
         s.s[1].cal.z = -s.s[1].raw.z;
      break;
   case SEL_XY:
      if (s.s[0].use)
         s.s[0].cal.x = -s.s[0].raw.x;
      if (s.s[1].use)
         s.s[1].cal.x = -s.s[1].raw.x;
      if (s.s[0].use)
         s.s[0].cal.y = -s.s[0].raw.y;
      if (s.s[1].use)
         s.s[1].cal.y = -s.s[1].raw.y;
      break;
   case SEL_YZ:
      if (s.s[0].use)
         s.s[0].cal.y = -s.s[0].raw.y;
      if (s.s[1].use)
         s.s[1].cal.y = -s.s[1].raw.y;
      if (s.s[0].use)
         s.s[0].cal.z = -s.s[0].raw.z;
      if (s.s[1].use)
         s.s[1].cal.z = -s.s[1].raw.z;
      break;
   case SEL_XZ:
      if (s.s[0].use)
         s.s[0].cal.x = -s.s[0].raw.x;
      if (s.s[1].use)
         s.s[1].cal.x = -s.s[1].raw.x;
      if (s.s[0].use)
         s.s[0].cal.z = -s.s[0].raw.z;
      if (s.s[1].use)
         s.s[1].cal.z = -s.s[1].raw.z;
      break;

   default:
      error(0, 0, "calibration not applicable");
      return;
   }
   save_cal();
}

static void cb_1d_axis (Fl_Widget *widget, long v) {
   int w, h, x, y;
   Fl_Group *parent = s.wid->parent();
   s.sel = v;
   w = parent->w();
   h = parent->h();
   switch (s.sel) {
   default:
   case SEL_X:
      h = 80;
      break;
   case SEL_Y:
   case SEL_Z:
      w = 80;
      break;
   }
   x = parent->x() + (parent->w() - w)/2;
   y = parent->y() + (parent->h() - h)/2;
   s.wid->type(FL_1D);
   s.wid->resize(x, y, w, h);
   parent->init_sizes();
   parent->redraw();
}
static void cb_2d_axis (Fl_Widget *widget, long v) {
   Fl_Group *parent = s.wid->parent();
   s.sel = v;
   s.wid->type(FL_2D);
   s.wid->resize(parent->x(), parent->y(), parent->w(), parent->h());
   parent->init_sizes();
   parent->redraw();
}

static void cb_sensor (Fl_Widget *widget, void * vp) {
   struct sens * s = (struct sens *)vp;
   //error(0, 0, "%s: %i", s->bttn->label(), s->bttn->value());
   s->use = s->bttn->value();
}

static void fix_presence (struct sens *s) {
   if ((time(0) - s->time) > 5) {
      s->use = 0;
      s->bttn->value(0);
   }
}

static void post_process (void) {
   int cnt;
   s.a.x = s.a.y = s.a.z = 0;
   cnt = 0;
   if (s.s[0].use) {
      s.a.x += s.s[0].raw.x + s.s[0].cal.x;
      s.a.y += s.s[0].raw.y + s.s[0].cal.y;
      s.a.z += s.s[0].raw.z + s.s[0].cal.z;
      ++cnt;
   }
   if (s.s[1].use) {
      s.a.x += s.s[1].raw.x + s.s[1].cal.x;
      s.a.y += s.s[1].raw.y + s.s[1].cal.y;
      s.a.z += s.s[1].raw.z + s.s[1].cal.z;
      ++cnt;
   }
   if (cnt > 1) {
      s.a.x /= cnt;
      s.a.y /= cnt;
      s.a.z /= cnt;
   }
   if (cnt) {
      s.wid->activate();
   } else {
      s.wid->deactivate();
   }
   switch (s.sel) {
   default:
   case SEL_X:
      s.wid->value(s.a.x);
      break;
   case SEL_Y:
      s.wid->value(s.a.y);
      break;
   case SEL_Z:
      s.wid->value(s.a.z);
      break;
   case SEL_XY:
      s.wid->value(s.a.x, s.a.y);
      break;
   case SEL_YZ:
      s.wid->value(s.a.y, s.a.z);
      break;
   case SEL_XZ:
      s.wid->value(s.a.x, s.a.z);
      break;
   }
}

static void on_timeout (void * p) {
   fix_presence(&s.s[0]);
   fix_presence(&s.s[1]);
   post_process();
   Fl::repeat_timeout(0.2, on_timeout, p);
}

static void on_input_event(int fd, void *vp) {
   struct sens * s = (struct sens *)vp;
   int ret;
   struct input_event * end;
   struct input_event * lp;
   struct input_event ev[4];

   ret = read(fd, ev, sizeof(ev));
   if (ret < 0) {
      if (EINTR == errno)
         return;
      error(1, errno, "read(%i)", fd);
      return;
   } else if (0 == ret) {
      return;
   }
   s->time = time(0);
   end = &ev[ret / sizeof(ev[0])];
   for (lp = ev; lp < end; ++lp) {
      switch (lp->type) {
      case EV_SYN:
         break;
      case EV_REL:
         switch (lp->code) {
         case REL_X:
            s->raw.x = lp->value /-1e3;
            break;
         case REL_Y:
            s->raw.y = lp->value /+1e3;
            break;
         case REL_Z:
            s->raw.z = lp->value /+1e3;
            break;
         }
         break;
      }
   }
}

static Fl_Button * mk_bttn (int x, int y, int w, int h, const char *l =0) {
   Fl_Button * wid;
   wid = new ::Fl_Light_Button(x, y, w, h, l);
   if (!wid)
      return 0;
   wid->align(FL_ALIGN_CENTER | FL_ALIGN_CLIP);
   wid->type(FL_RADIO_BUTTON);
   wid->when(FL_WHEN_CHANGED);
   wid->selection_color(FL_MAGENTA);
   wid->color(fl_rgb_color(0x40, 0x40, 0x40));
   return wid;
}

int main (int argc, char *argv[]) {
   int fd[2];

   s.s[0].cal.x = s.s[0].cal.y = s.s[0].cal.z = 0;
   s.s[1].cal.x = s.s[1].cal.y = s.s[1].cal.z = 0;

   load_cal();

   error(0, 0, "using %s & %s", dev0, dev1);

   fd[0] = open(dev0, O_RDONLY);
   if (fd[0] < 0)
      error(1, errno, "open(%s)", dev0);
   fd[1] = open(dev1, O_RDONLY);
   if (fd[1] < 0)
      error(1, errno, "open(%s)", dev1);

   FL_NORMAL_SIZE = 24;
   ::Fl::get_system_colors();
   ::Fl::foreground(0xff, 0xff, 0xff);
   ::Fl::background(0x00, 0x00, 0x00);
   ::Fl::visible_focus(0);
   int texth = FL_NORMAL_SIZE *5/3;
   int x, y, w, h;
   x = y = 0;
   w = ::Fl::w();
   h = ::Fl::h();
   ::Fl_Window *dlg = new ::Fl_Double_Window(w, h, "gwaterpas");
   dlg->resizable(dlg);
   x += 10;
   y += 10;
   w -= 20;
   h -= 20;
   int bw = w /4;
   ::Fl_Button *bttn;

   if ((0 != (bttn = mk_bttn(x + 0*bw, y, bw, texth, "X")))) {
      bttn->callback(cb_1d_axis, (long)SEL_X);
   }
   if ((0 != (bttn = mk_bttn(x + 1*bw, y, bw, texth, "Y")))) {
      bttn->callback(cb_1d_axis, (long)SEL_Y);
      bttn->value(1);
   }
   if ((0 != (bttn = mk_bttn(x + 2*bw, y, bw, texth, "Z")))) {
      bttn->callback(cb_1d_axis, (long)SEL_Z);
   }
   if ((0 != (s.s[0].bttn = bttn = mk_bttn(x + 3*bw, y, bw/2, texth, "s0")))) {
      bttn->type(FL_TOGGLE_BUTTON);
      bttn->value(1);
      s.s[1].use = 1;
      bttn->callback(cb_sensor, &s.s[0]);
   }
   if ((0 != (s.s[1].bttn = bttn = mk_bttn(x + 3*bw +bw/2, y, bw - bw/2, texth, "s1")))) {
      bttn->type(FL_TOGGLE_BUTTON);
      bttn->value(1);
      s.s[1].use = 1;
      bttn->callback(cb_sensor, &s.s[1]);
   }
   y += texth;
   h -= texth;
   if ((0 != (bttn = mk_bttn(x + 0*bw, y, bw, texth, "XY")))) {
      bttn->callback(cb_2d_axis, (long)SEL_XY);
   }
   if ((0 != (bttn = mk_bttn(x + 1*bw, y, bw, texth, "YZ")))) {
      bttn->callback(cb_2d_axis, (long)SEL_YZ);
   }
   if ((0 != (bttn = mk_bttn(x + 2*bw, y, bw, texth, "XZ")))) {
      bttn->callback(cb_2d_axis, (long)SEL_XZ);
   }
   if ((0 != (bttn = mk_bttn(x + 3*bw, y, bw, texth, "calibrate")))) {
      s.cal_bttn = bttn;
      bttn->type(0);
      bttn->when(FL_WHEN_RELEASE);
      bttn->callback(cb_calibrate, bttn);
      bttn->labelsize(16);
   }
   y += texth +4;
   h -= texth +4;
   ::Fl_Group *grp;
   grp = new ::Fl_Group(x, y, w, h);
   if (grp) {
      s.wid = new Fl_Level(x, y, w, h/2);
      //s.wid->range(-1e3, 1e3);
      grp->box(FL_FLAT_BOX);
      grp->end();
      grp->resizable(s.wid);
   }
   dlg->resizable(grp);
   dlg->end();
   cb_1d_axis(0, SEL_Y);
   dlg->show();
   ::Fl::add_timeout(0.1, on_timeout, 0);
   ::Fl::add_fd(fd[0], on_input_event, &s.s[0]);
   ::Fl::add_fd(fd[1], on_input_event, &s.s[1]);
   ::Fl::run();
   ::Fl::flush();
   return 0;
}

static void save_cal (void) {
   FILE *fp;
   char * file;
   error(0, 0, "saving cal");
   asprintf(&file, "%s/.gwaterpas", getenv("HOME"));
   fp = fopen(file, "w");
   if (!fp) {
      error(0, errno, "fopen(%s)", file);
      free(file);
      return;
   }
   free(file);
   fprintf(fp, "x[0]\t%.6lf\n", s.s[0].cal.x);
   fprintf(fp, "y[0]\t%.6lf\n", s.s[0].cal.y);
   fprintf(fp, "z[0]\t%.6lf\n", s.s[0].cal.z);
   fprintf(fp, "x[1]\t%.6lf\n", s.s[1].cal.x);
   fprintf(fp, "y[1]\t%.6lf\n", s.s[1].cal.y);
   fprintf(fp, "z[1]\t%.6lf\n", s.s[1].cal.z);
   fclose(fp);
}
static void load_cal (void) {
   static const char ws [] = " \t\r\n\f\v";
   FILE *fp;
   char *line;
   size_t size;
   int ret;
   char * name;
   char * tok;
   double val;
   char * file;
   int linenr;

   asprintf(&file, "%s/.gwaterpas", getenv("HOME"));
   fp = fopen(file, "r");
   if (!fp) {
      error(0, errno, "fopen(%s)", file);
      free(file);
      return;
   }
   error(0, 0, "loading cal from %s", file);
   free(file);

   line = 0;
   size = 0;
   linenr = 0;
   while ((ret = getline(&line, &size, fp)) > 0) {
      ++linenr;
      if (line[0] == '#')
         continue;
      name = strtok(line, ws);
      if (!name)
         continue;
      tok = strtok(0, ws);
      if (!tok)
         continue;
      val = strtod(tok, 0);

      if (!strcmp("x[0]", name))
         s.s[0].cal.x = val;
      else if (!strcmp("y[0]", name))
         s.s[0].cal.y = val;
      else if (!strcmp("z[0]", name))
         s.s[0].cal.z = val;
      else if (!strcmp("x[1]", name))
         s.s[1].cal.x = val;
      else if (!strcmp("y[1]", name))
         s.s[1].cal.y = val;
      else if (!strcmp("z[1]", name))
         s.s[1].cal.z = val;
      else
         error(0, 0, "line %u: token '%s' not found", linenr, name);
   }
   fclose(fp);
}

